// Author:Richardson, Zhu
// Date: 11-29-2006
// split SV testbench to writer and reader

// version2 to version3
// add @(posedge clk) to all tasks to guarantee data be set

module top;


parameter bit ACTIVE_RESET = 0;
typedef int FILE;

parameter WIDTH = 32;
typedef logic [WIDTH-1 : 0] word_t;
typedef bit [1:0] idle_dly_t;

//declear queue for scoreboard
word_t before_DUT_queue[$:20];
word_t after_DUT_queue[$:20];

//interface instance here
sync_fifo_if sync_fifo_if_i();

typedef enum { READ = 0, RD_IDLE = 1} rd_cmd_t;
typedef enum { WRITE = 0, WR_IDLE = 1 } wr_cmd_t;

// system reset: not implemented yet
// typedef enum { RESET = 0, NORMAL = 1 } sys_cmd_t;

//----------------------
// transaction stimulus
//----------------------
class wr_tr_c;

    rand wr_cmd_t wr_cmd;
    rand word_t data;
    rand idle_dly_t idle_dly; // idle delay is between 1~3 cycles
    
        constraint wr_cmd_range {
            wr_cmd dist { WRITE := 4, WR_IDLE := 1 };
        }

        constraint data_range {
            //data < 20;
            data != 0; // when receive 0 data, there must be error
        }

        constraint dly_range {
            idle_dly != 0;
        }
        

endclass:wr_tr_c

class rd_tr_c;
    
    rand rd_cmd_t rd_cmd;
    rand idle_dly_t idle_dly; // idle delay is between 1~3 cycles
    word_t data;

        constraint rd_cmd_range {
            rd_cmd dist { READ := 4, RD_IDLE := 1 };
        }

        constraint dly_range {
            idle_dly != 0;
        }
        
endclass:rd_tr_c

//--------------------
// transaction driver
//--------------------
class wr_driver_c;
    
    //declear interface modport (5.81 not support virtual interface)
    //sync_fifo_if sync_fifo_if_i;
    
    //declear transaction
    wr_tr_c wr_tr;
    word_t wr_data;
    int random_type;
    //string report_str;
    
    task run;
    begin
        $display ("wr_driver start!");
        forever begin
            $strobe("G0.1 @%0t sync_fifo_if_i.wr is %d", $time, sync_fifo_if_i.wr);
            @ (sync_fifo_if_i.wr_driver_cb);
            begin
                if (sync_fifo_if_i.rst == ACTIVE_RESET) begin
                    //reset sync_fifo_rtl
                    sync_fifo_if_i.wr_driver_cb.wr <= 1'b0;
                    sync_fifo_if_i.wr_driver_cb.din <= 0;              
                    continue;
                end
                
                wr_tr = new();
                random_type = wr_tr.randomize();
                
                if (wr_tr.wr_cmd == WRITE) begin
                    wr_data = wr_tr.data;
                    $display ("@%0t wr_data is %d",$time, wr_data);
                    do_write();
                    $display ("@%0t *WRITE cmd",$time);
                    wr_data = wr_tr.data;
                    $display ("@%0t wr_data is %d", $time, wr_data);
                    $display ("@%0t sync_fifo_if_i.din is %d",$time ,sync_fifo_if_i.din);

                end
                else begin
                    idle();
                    $display ("@%0t *WR_IDLE cmd",$time);
                end
            end
        end
    end
    endtask:run

    task do_write;
    begin
        $display ("@%0t Enter W::do_write", $time);
        //if ((sync_fifo_if_i.wr_driver_cb.almost_full == 0) && (sync_fifo_if_i.wr_driver_cb.full == 0)) begin
        if (sync_fifo_if_i.wr_driver_cb.almost_full == 0) begin
            sync_fifo_if_i.wr_driver_cb.din <= wr_tr.data;
            sync_fifo_if_i.wr_driver_cb.wr <= 1'b1;
        end
        else begin //fifo almost_full 
            sync_fifo_if_i.wr_driver_cb.din <= '0;
            sync_fifo_if_i.wr_driver_cb.wr <= '0;
            
            // below $display is illegal: clock block output cannot be rvalue
            // $display("@%0t wr_driver_cb.wr is %d", $time, sync_fifo_if_i.wr_driver_cb.wr);
        end
        $strobe("F1 @%0t sync_fifo_if_i.wr is %d", $time, sync_fifo_if_i.wr);
        $display ("@%0t Quit W::do_write", $time);
    end
    endtask:do_write

    task idle;
    begin
        $display ("Enter W::idle");
        sync_fifo_if_i.wr_driver_cb.din <= 0;
        sync_fifo_if_i.wr_driver_cb.wr <= 1'b0;
        repeat (wr_tr.idle_dly) @(sync_fifo_if_i.wr_driver_cb);
        $display ("Quit W::idle");
    end
    endtask:idle
    
endclass:wr_driver_c

class rd_driver_c;
    
    //declear interface modport (5.81 not support virtual interface)
    //sync_fifo_if sync_fifo_if_i;
    
    //declear transaction
    rd_tr_c rd_tr;
    
    int random_type;
    task run;
    begin
        $display ("rd_driver start!");
        forever begin
            @ (sync_fifo_if_i.rd_driver_cb)
            if (sync_fifo_if_i.rst == ACTIVE_RESET) begin
                //reset sync_fifo_rtl
                sync_fifo_if_i.rd_driver_cb.rd <= 1'b0;
                continue;
            end
            
            rd_tr = new();
            random_type = rd_tr.randomize();
            //if (rd_tr.randomize()) 
            //    $finish();
            
            if (rd_tr.rd_cmd == READ) begin
                do_read();
                $display ("@%0t *READ cmd",$time);
            end
            else begin
                idle();
                $display ("@%0t *RD_IDLE cmd",$time);
            end
        end
    end
    endtask:run

    task do_read;
    begin
        $display ("Enter R::do_read");
        if (sync_fifo_if_i.rd_driver_cb.empty == 0) begin
            sync_fifo_if_i.rd_driver_cb.rd <= 1'b1;
        end
        else begin
            $display("@%0t empty == 1", $time);
            sync_fifo_if_i.rd_driver_cb.rd <= 1'b0;
        end
        $strobe("@%0t sync_fifo_if_i.rd is %d", $time, sync_fifo_if_i.rd);
        $display ("Quit R::do_read");
    end
    endtask:do_read

    task idle;
    begin
        $display ("Enter R::idle");
        sync_fifo_if_i.rd_driver_cb.rd <= 1'b0;
        repeat (rd_tr.idle_dly) @(sync_fifo_if_i.rd_driver_cb);
        $display ("Quit R::idle");
    end
    endtask:idle
    
endclass:rd_driver_c
 
//--------------------
// reporter
//--------------------
//class reporter;

// NC5.82 does not support string yet

//  static function void f_display( FILE f , string s );
//    //f==0 means screen display
//    if( f == 0 ) begin
//      $display( s );
//    end
//    else begin
//      $fdisplay( f , s );
//    end
//  endfunction
  
//endclass:reporter

//----------------------
// link DUT and testbench to interface 
// fifo instance here
//----------------------
/*
sync_fifo_rtl sync_fifo_rtl_i(
    .CLK (sync_fifo_if_i.clk),
    .RESET (sync_fifo_if_i.rst),
    
    // control signal
    .RD (sync_fifo_if_i.rd),
    .WR (sync_fifo_if_i.wr),
    // data path
    .DIN (sync_fifo_if_i.din),
    .DOUT(sync_fifo_if_i.dout),
    // status signal
    .FULL (sync_fifo_if_i.full),
    .EMPTY (sync_fifo_if_i.empty),
    .ALMOST_FULL (sync_fifo_if_i.almost_full), // one clock ahead of FULL
    .ALMOST_EMPTY (sync_fifo_if_i.almost_empty) // one clock ahead of EMPTY
);
*/

sync_fifo_rtl sync_fifo_rtl_i(
    .CLK (sync_fifo_if_i.sync_fifo_rtl_mp.clk),
    .RESET (sync_fifo_if_i.sync_fifo_rtl_mp.rst),
    
    // control signal
    .RD (sync_fifo_if_i.sync_fifo_rtl_mp.rd),
    .WR (sync_fifo_if_i.sync_fifo_rtl_mp.wr),
    // data path
    .DIN (sync_fifo_if_i.sync_fifo_rtl_mp.din),
    .DOUT(sync_fifo_if_i.sync_fifo_rtl_mp.dout),
    // status signal
    .FULL (sync_fifo_if_i.sync_fifo_rtl_mp.full),
    .EMPTY (sync_fifo_if_i.sync_fifo_rtl_mp.empty),
    .ALMOST_FULL (sync_fifo_if_i.sync_fifo_rtl_mp.almost_full), // one clock ahead of FULL
    .ALMOST_EMPTY (sync_fifo_if_i.sync_fifo_rtl_mp.almost_empty) // one clock ahead of EMPTY
);

//------------------------------------------
// two monitor: writer side and reader side
//------------------------------------------
class wr_monitor_c;
    //wr_tr_c wr_tr;
    
    task run;
        $display ("wr_monitor start!");
        forever begin
            @ (sync_fifo_if_i.wr_monitor_cb);
            if (sync_fifo_if_i.wr_monitor_cb.wr == 1) begin
                before_DUT_queue.push_back(sync_fifo_if_i.wr_monitor_cb.din);
                $display ("@%0t before DUT, push %d", $time, sync_fifo_if_i.wr_monitor_cb.din);
            end
        end
    endtask:run
    
endclass:wr_monitor_c
   

class rd_monitor_c;
    //rd_tr_c rd_tr;
    local bit get_dout; //indicate whether to sample rd_monitor_cb.dout and the next posedge clk
    task run;
        $display ("rd_monitor start!");
        forever begin
            
            @ (sync_fifo_if_i.rd_monitor_cb);
            
            if (get_dout == 1) begin
                after_DUT_queue.push_back(sync_fifo_if_i.rd_monitor_cb.dout);
                $display ("@%0t after DUT, push %d", $time, sync_fifo_if_i.rd_monitor_cb.dout);
            end
            
            if (sync_fifo_if_i.rd_monitor_cb.rd == 1) begin
                get_dout = 1;
            end
            else begin
                get_dout = 0;
            end
        end
    endtask:run
endclass:rd_monitor_c

//----------------------------
// scoreboard: queue and comparator
//----------------------------
class scoreboard;
    
    task run;
        word_t sb_din,sb_dout;
        $display ("scoreboard start!");
        forever begin
            @ (sync_fifo_if_i.sb_cb);
            if ( (after_DUT_queue.size() != 0) && (before_DUT_queue.size() != 0) ) begin
                //compare data in before_DUT_queue and after_DUT_queue
                //compare(after_DUT_queue.pop_front(),before_DUT_queue.pop_front());
                sb_din = before_DUT_queue.pop_front();
                sb_dout = after_DUT_queue.pop_front();
                
                if (sb_dout == sb_din) begin
                    $display ("@%0t dout is %0d, din is %0d, pass!", $time, sb_dout, sb_din);
                end
                else begin
                    $display ("@%0t dout is %0d, din is %0d, fail!", $time, sb_dout, sb_din );
                end
            end
        end
        
    endtask:run
/*
    void function compare(input word_t dout, input word_t din);
        if (dout == din) begin
            $display ("@%0t dout is %d, din is %d, pass!");
        end
        else begin
            $display ("@%0t dout is %d, din is %d, fail!");
        end
    endfunction:compare
*/    
endclass:scoreboard

//--------------------
// main run thread
//--------------------
// initialize the testbench module
clock_reset cr (sync_fifo_if_i);
wr_driver_c wr_driver;
rd_driver_c rd_driver;
//wr_monitor_c wr_monitor;
//rd_monitor_c rd_monitor;
//scoreboard sb;
FILE f_report;

initial begin
    $display ("initial start!\n");
    wr_driver = new;
    rd_driver = new;
    //wr_monitor = new;
    //rd_monitor = new;
    fork
        $display("fork start!\n");
        cr.run;
        
        wr_driver.run;
        rd_driver.run;
        
        //wr_monitor.run;
        //rd_monitor.run;

        //sb.run;
        
    join_none

    #3000 $finish;
end

endmodule:top

